/**-*-C++-*-
 * \file
 *
 * \brief Style definition.
 *
 * This file is distributed under the GNU General Public License. See
 * COPYING for details.
 */
#ifndef IRONY_MODE_SERVER_STYLE_H_
#define IRONY_MODE_SERVER_STYLE_H_

enum class PrefixMatchStyle {
  Exact, CaseInsensitive, SmartCase,
};


#endif //IRONY_MODE_SERVER_STYLE_H_
